import numpy as np
import pandas as pd
import pytest

from opti.constraint import LinearEquality
from opti.objective import CloseToTarget, Maximize, Minimize
from opti.parameter import Continuous
from opti.problem import Problem
from opti.problems import Cake, Photodegradation, Zakharov_Constrained
from opti.tools import sanitize_problem


@pytest.mark.filterwarnings("ignore:f is not sanitized but dropped")
def test_sanitize_problem():
    def _test(original):
        sanitized = sanitize_problem(original)

        # check inputs
        for p1, p2 in zip(original.inputs, sanitized.inputs):
            assert type(p1) == type(p2)
            if isinstance(p2, Continuous):
                p2.bounds == (0, 1)  # continuous input bounds are [0, 1]
            assert p2.name.startswith("input")

        # check outputs
        for p1, p2 in zip(original.outputs, sanitized.outputs):
            assert type(p1) == type(p2)
            if isinstance(
                p2, Continuous
            ):  # continuous output bounds are [0, 1] if specified and [-inf, inf] if not
                if np.isneginf(p1.low):
                    assert np.isneginf(p2.low)
                else:
                    assert np.isclose(p2.low, 0)
                if np.isinf(p1.high):
                    assert np.isinf(p2.high)
                else:
                    assert np.isclose(p2.high, 1)
            assert p2.name.startswith("output")

        # check objectives
        for (p1, p2) in zip(original.objectives, sanitized.objectives):
            assert type(p1) == type(p2)
            assert p2.name.startswith("output")
            assert np.isfinite(p2.target)

        # check constraints
        if original.constraints is not None:
            for c1, c2 in zip(original.constraints, sanitized.constraints):
                # sanitizing should not modify the original problem
                assert c1.names != c2.names

        # check data
        assert original.data.shape == sanitized.data.shape
        assert set(sanitized.data.columns) == set(
            sanitized.inputs.names + sanitized.outputs.names
        )
        assert sanitized.inputs.contains(sanitized.data[sanitized.inputs.names]).all()
        assert sanitized.constraints.satisfied(sanitized.data).all()

        # problem.f is dropped
        assert not hasattr(sanitized, "f")

    _test(Cake())

    _test(Photodegradation())

    z = Zakharov_Constrained()
    z.create_initial_data(n_samples=1000)
    _test(z)

    tp = Problem(
        inputs=[
            Continuous("secret_ingredient_1", domain=[0, 150]),
            Continuous("secret_ingredient_2", domain=[0, 50]),
            Continuous("secret_ingredient_3", domain=[0, 34.3]),
        ],
        outputs=[
            Continuous("ultimate_goal", domain=[0, np.linalg.norm([150, 50, 34.3])]),
            Continuous("mega_goal", domain=[0, np.sum(np.square([150, 50, 34.3]))]),
            Continuous(
                "super_goal", domain=[0, np.linalg.norm([150, 50, 34.3]) ** 0.5]
            ),
        ],
        objectives=[
            CloseToTarget("ultimate_goal", target=100),
            Minimize("mega_goal", target=1000),
            Maximize("super_goal", target=10),
        ],
        constraints=[
            LinearEquality(
                ["secret_ingredient_1", "secret_ingredient_2", "secret_ingredient_3"],
                lhs=np.array([3.5, 4.5, 2]),
                rhs=153.1,
            )
        ],
        f=lambda df: pd.DataFrame(
            {
                "ultimate_goal": np.linalg.norm(df.to_numpy(), axis=1),
                "mega_goal": np.sum(df**2, axis=1),
                "super_goal": np.linalg.norm(df, axis=1) ** 0.5,
            }
        ),
    )
    tp.create_initial_data(n_samples=100)
    _test(tp)

    empty = Problem([], [])  # this should not be possible
    with pytest.raises(TypeError):
        sanitize_problem(empty)



import numpy as np
import pytest

import opti
from opti.parameter import Continuous, Discrete


def sample_and_check_function(problem):
    X = problem.sample_inputs(10)
    X.index += 1000  # change index to test wether it is kept
    Y = problem.f(X)
    assert np.all(X.index == Y.index)
    assert np.all(problem.outputs.contains(Y))


def test_single_objective_problems():
    for _Problem in (
        opti.problems.Ackley,
        opti.problems.Branin,
        opti.problems.Himmelblau,
        opti.problems.Michalewicz,
        opti.problems.Rastrigin,
        opti.problems.Rosenbrock,
        opti.problems.Schwefel,
        opti.problems.Sphere,
        opti.problems.ThreeHumpCamel,
        opti.problems.Zakharov,
        opti.problems.Zakharov_Categorical,
        opti.problems.Zakharov_Constrained,
        opti.problems.Zakharov_NChooseKConstraint,
    ):
        problem = _Problem()
        sample_and_check_function(problem)
        optima = problem.get_optima()
        px = optima[problem.inputs.names]
        py = optima[problem.outputs.names]
        assert np.allclose(problem.f(px), py, atol=1e-04)


def test_multi_objective_problems():
    for _Problem in (
        opti.problems.Daechert1,
        opti.problems.Daechert2,
        opti.problems.Daechert3,
        opti.problems.Hyperellipsoid,
        opti.problems.OmniTest,
        opti.problems.Poloni,
        opti.problems.Qapi1,
        opti.problems.WeldedBeam,
    ):
        problem = _Problem()
        sample_and_check_function(problem)


def test_dataset_problems():
    for _Problem in (
        opti.problems.Alkox,
        opti.problems.BaumgartnerAniline,
        opti.problems.BaumgartnerBenzamide,
        opti.problems.Benzylation,
        opti.problems.Cake,
        opti.problems.Fullerenes,
        opti.problems.HPLC,
        opti.problems.Photodegradation,
        opti.problems.ReizmanSuzuki,
        opti.problems.Suzuki,
        opti.problems.SnAr,
    ):
        problem = _Problem()
        data = problem.get_data()
        assert problem.inputs.contains(data).all()
        # assert problem.outputs.contains(data).all()


def test_hyperellipsoid_problem():
    # Hypersphere
    problem = opti.problems.Hyperellipsoid(n=3)
    optima = problem.get_optima(10)
    assert len(optima) == 10

    # Hyperellipsoid
    with pytest.raises(ValueError):
        opti.problems.Hyperellipsoid(n=3, a=[1, 1])  # a has wrong shape

    problem = opti.problems.Hyperellipsoid(n=2, a=[1000, 1])
    sample_and_check_function(problem)

    optima = problem.get_optima(10)
    optima.loc[0, "y1"] = -1000
    optima.loc[0, "y2"] = -1


def test_detergent():
    problem = opti.problems.Detergent()
    sample_and_check_function(problem)

    problem = opti.problems.Detergent_OutputConstraint(discrete=False)
    assert isinstance(problem.outputs["stable"], Continuous)
    sample_and_check_function(problem)

    problem = opti.problems.Detergent_OutputConstraint(discrete=True)
    assert isinstance(problem.outputs["stable"], Discrete)
    sample_and_check_function(problem)

    problem = opti.problems.Detergent_TwoOutputConstraints()
    X = problem.sample_inputs(1000)
    Y = problem.f(X)
    # test that y1 - y5 is only available if stability 1 = 1
    cols = [f"y{i+1}" for i in range(5)]
    assert np.all(Y[cols][Y["stability 1"] == 0].isna())
    assert np.all(Y[cols][Y["stability 1"] == 1].notna())

    problem = opti.problems.Detergent_NChooseKConstraint()
    # problem.create_initial_data(10)  # sampling for n-choose-k constraints not implemented


def test_zdt_problems():
    for _Problem in (
        opti.problems.ZDT1,
        opti.problems.ZDT2,
        opti.problems.ZDT3,
        opti.problems.ZDT4,
        opti.problems.ZDT6,
    ):
        problem = _Problem(n_inputs=10)
        sample_and_check_function(problem)
        optima = problem.get_optima(50)
        assert len(optima) == 50


def test_univariate_problems():
    for _Problem in (
        opti.problems.Line1D,
        opti.problems.Parabola1D,
        opti.problems.Sinus1D,
        opti.problems.Sigmoid1D,
        opti.problems.Step1D,
    ):
        problem = _Problem()
        sample_and_check_function(problem)


def test_mixed_variables_problems():
    problem = opti.problems.DiscreteVLMOP2(5)
    x = problem.inputs.sample(1000)
    y = problem.f(x)
    assert np.all(problem.outputs.contains(y))

    problem = opti.problems.DiscreteFuelInjector()
    x = problem.inputs.sample(1000)
    y = problem.f(x)
    assert np.all(problem.outputs.contains(y))



"""
Chemical datasets.
These problems contain observed data but don't come with a ground truth.
"""
import os

import pandas as pd

from opti.constraint import LinearEquality, NChooseK
from opti.objective import CloseToTarget, Maximize, Minimize
from opti.parameter import Categorical, Continuous
from opti.problem import Problem

cdir = os.path.dirname(os.path.realpath(__file__))


def get_data(fname: str):
    return pd.read_csv(f"{cdir}/data/{fname}")


class Cake(Problem):
    """Cake recipe optimization with mixed objectives."""

    def __init__(self):
        super().__init__(
            name="Cake",
            inputs=[
                Continuous("wheat_flour", domain=[0, 1]),
                Continuous("spelt_flour", domain=[0, 1]),
                Continuous("sugar", domain=[0, 1]),
                Continuous("chocolate", domain=[0, 1]),
                Continuous("nuts", domain=[0, 1]),
                Continuous("carrot", domain=[0, 1]),
            ],
            outputs=[
                Continuous("calories", domain=[300, 600]),
                Continuous("taste", domain=[0, 5]),
                Continuous("browning", domain=[0, 2]),
            ],
            objectives=[
                Minimize("calories"),
                Maximize("taste"),
                CloseToTarget("browning", target=1.4),
            ],
            constraints=[
                LinearEquality(
                    [
                        "wheat_flour",
                        "spelt_flour",
                        "sugar",
                        "chocolate",
                        "nuts",
                        "carrot",
                    ],
                    rhs=1,
                )
            ],
            data=get_data("cake.csv"),
        )


class Alkox(Problem):
    """Alkoxylation dataset

    This dataset reports the biocatalytic oxidation of benzyl alcohol by a copper radical oxidase (AlkOx).
    The effects of enzyme loading, cocatalyst loading, and pH balance on both initial rate and total conversion were assayed.
    The dataset includes 104 samples with four parameters and one objective.

    Reference:
        F. Häse, M. Aldeghi, R.J. Hickman, L.M. Roch, M. Christensen, E. Liles, J.E. Hein, A. Aspuru-Guzik. Olympus: a benchmarking framework for noisy optimization and experiment planning. arXiv (2020), 2010.04153.
        [DOI](https://doi.org/10.1088/2632-2153/abedc8).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="Alkox",
            inputs=[
                Continuous("residence_time", domain=[0.05, 1]),
                Continuous("ratio", domain=[0.5, 10]),
                Continuous("concentration", domain=[2, 8]),
                Continuous("temperature", domain=[6, 8]),
            ],
            outputs=[Continuous("conversion")],
            objectives=[Maximize("conversion")],
            data=get_data("alkox.csv"),
        )


class BaumgartnerAniline(Problem):
    """Aniline C-N cross-coupling dataset.

    Reference:
        Baumgartner et al. 2019 - Use of a Droplet Platform To Optimize Pd-Catalyzed C-N Coupling Reactions Promoted by Organic Bases
        [DOI](https://doi.org/10.1021/acs.oprd.9b00236).
        Data obtained from [Summit](https://github.com/sustainable-processes/summit).
    """

    def __init__(self):
        super().__init__(
            name="Aniline cross-coupling, Baumgartner 2019",
            inputs=[
                Categorical("catalyst", domain=["tBuXPhos", "tBuBrettPhos", "AlPhos"]),
                Categorical("base", domain=["TEA", "TMG", "BTMG", "DBU"]),
                Continuous("base_equivalents", domain=[1.0, 2.5]),
                Continuous("temperature", domain=[30, 100]),
                Continuous("residence_time", domain=[60, 1800]),
            ],
            outputs=[Continuous("yield", domain=[0, 1])],
            objectives=[Maximize("yield")],
            data=get_data("baumgartner_aniline.csv"),
        )


class BaumgartnerBenzamide(Problem):
    """Benzamide C-N cross-coupling dataset.

    Reference:
        Baumgartner et al. 2019 - Use of a Droplet Platform To Optimize Pd-Catalyzed C-N Coupling Reactions Promoted by Organic Bases
        [DOI](https://doi.org/10.1021/acs.oprd.9b00236).
        Data obtained from [Summit](https://github.com/sustainable-processes/summit).
    """

    def __init__(self):
        super().__init__(
            name="Benzamide cross-coupling, Baumgartner 2019",
            inputs=[
                Categorical("catalyst", domain=["tBuXPhos", "tBuBrettPhos"]),
                Categorical("base", domain=["TMG", "BTMG", "DBU", "MTBD"]),
                Continuous("base_equivalents", domain=[1.0, 2.1]),
                Continuous("temperature", domain=[30, 100]),
                Continuous("residence_time", domain=[60, 1850]),
            ],
            outputs=[Continuous("yield", domain=[0, 1])],
            objectives=[Maximize("yield")],
            data=get_data("baumgartner_benzamide.csv"),
        )


class Benzylation(Problem):
    """Benzylation dataset.

    This dataset reports the yield of undesired product (impurity) in an N-benzylation reaction.
    Four conditions of this reaction performed in a flow reactor can be controlled to minimize the yield of impurity.
    The dataset includes 73 samples with four parameters and one objective.

    Reference:
        A.M. Schweidtmann, A.D. Clayton, N. Holmes, E. Bradford, R.A. Bourne, A.A. Lapkin. Machine learning meets continuous flow chemistry: Automated optimization towards the Pareto front of multiple objectives. Chem. Eng. J. 352 (2018) 277-282.
        [DOI](https://doi.org/10.1016/j.cej.2018.07.031).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="Benzylation",
            inputs=[
                Continuous("flow_rate", domain=[0.2, 0.4]),
                Continuous("ratio", domain=[1.0, 5.0]),
                Continuous("solvent", domain=[0.5, 1.0]),
                Continuous("temperature", domain=[110.0, 150.0]),
            ],
            outputs=[Continuous("impurity")],
            objectives=[Minimize("impurity")],
            data=get_data("benzylation.csv"),
        )


class Fullerenes(Problem):
    """Buckminsterfullerene dataset.

    This dataset reports the production of o-xylenyl adducts of Buckminsterfullerenes.
    Three process conditions (temperature, reaction time and ratio of sultine to C60) are varied to maximize the mole fraction of the desired product.
    Experiments are executed on a three factor fully factorial grid with six levels per factor.
    The dataset includes 246 samples with three parameters and one objective.

    Reference:
        B.E. Walker, J.H. Bannock, A.M. Nightingale, J.C. deMello. Tuning reaction products by constrained optimisation. React. Chem. Eng., (2017), 2, 785-798.
        [DOI](https://doi.org/10.1039/C7RE00123A).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="Fullerenes",
            inputs=[
                Continuous("reaction_time", domain=[3.0, 31.0]),
                Continuous("sultine", domain=[1.5, 6.0]),
                Continuous("temperature", domain=[100.0, 150.0]),
            ],
            outputs=[Continuous("product")],
            objectives=[Maximize("product")],
            data=get_data("fullerenes.csv"),
        )


class HPLC(Problem):
    """High-performance liquid chromatography dataset.

    This dataset reports the peak response of an automated high-performance liquid chromatography (HPLC) system for varying process parameters.
    The dataset includes 1,386 samples with six parameters and one objective.

    Reference:
        L.M. Roch, F. Häse, C. Kreisbeck, T. Tamayo-Mendoza, L.P.E. Yunker, J.E. Hein, A. Aspuru-Guzik. ChemOS: an orchestration software to democratize autonomous discovery. (2018)
        [DOI](https://doi.org/10.26434/chemrxiv.5953606.v1).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="HPLC",
            inputs=[
                Continuous("sample_loop", domain=[0.0, 0.08]),
                Continuous("additional_volume", domain=[0.0, 0.06]),
                Continuous("tubing_volume", domain=[0.1, 0.9]),
                Continuous("sample_flow", domain=[0.5, 2.5]),
                Continuous("push_speed", domain=[80.0, 150]),
                Continuous("wait_time", domain=[0.5, 10.0]),
            ],
            outputs=[Continuous("peak_area")],
            objectives=[Maximize("peak_area")],
            data=get_data("hplc.csv"),
        )


class Photodegradation(Problem):
    """Photodegration dataset.

    This dataset reports the degradation of polymer blends for organic solar cells under the exposure to light.
    Individual data points encode the ratios of individual polymers in one blend, along with the measured photodegradation of this blend.
    The dataset includes 2,080 samples with five parameters and one objective.

    Reference:
        S. Langner, F. Häse, J.D. Perea, T. Stubhan, J. Hauch, L.M. Roch, T. Heumueller, A. Aspuru-Guzik, C.J. Brabec. Beyond Ternary OPV: High-Throughput Experimentation and Self-Driving Laboratories Optimize Multicomponent Systems. Advanced Materials, 2020, 1907801.
        [DOI](https://doi.org/10.1002/adma.201907801).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="Photodegradation",
            inputs=[
                Continuous("PCE10", domain=[0, 1]),
                Continuous("WF3", domain=[0, 1]),
                Continuous("P3HT", domain=[0, 1]),
                Continuous("PCBM", domain=[0, 1]),
                Continuous("oIDTBR", domain=[0, 1]),
            ],
            outputs=[Continuous("degradation")],
            objectives=[Minimize("degradation")],
            constraints=[
                LinearEquality(
                    ["PCE10", "WF3", "P3HT", "PCBM", "oIDTBR"], rhs=1, lhs=1
                ),
                NChooseK(["PCE10", "WF3"], max_active=1),
            ],
            data=get_data("photodegradation.csv"),
        )


class ReizmanSuzuki(Problem):
    """Suzuki-Miyaura cross-coupling optimization.

    Each case was has a different set of substrates but the same possible catalysts.

    Reference:
        Reizman et al. (2016) Suzuki-Miyaura cross-coupling optimization enabled by automated feedback. Reaction chemistry & engineering, 1(6), 658-666
        [DOI](https://doi.org/10.1039/C6RE00153J).
        Data obtained from [Summit](https://github.com/sustainable-processes/summit).
    """

    def __init__(self, case=1):
        assert case in [1, 2, 3, 4]
        super().__init__(
            name=f"Reizman 2016 - Suzuki Case {case}",
            inputs=[
                Categorical(
                    "catalyst",
                    domain=[
                        "P1-L1",
                        "P2-L1",
                        "P1-L2",
                        "P1-L3",
                        "P1-L4",
                        "P1-L5",
                        "P1-L6",
                        "P1-L7",
                    ],
                ),
                Continuous("t_res", domain=[60, 600]),
                Continuous("temperature", domain=[30, 110]),
                Continuous("catalyst_loading", domain=[0.496, 2.515]),
            ],
            outputs=[
                Continuous("ton", domain=[0, 100]),
                Continuous("yield", domain=[0, 100]),
            ],
            objectives=[Maximize("ton"), Maximize("yield")],
            data=get_data(f"reizman_suzuki{case}.csv"),
        )


class SnAr(Problem):
    """SnAr reaction optimization.

    This dataset reports the e-factor for a nucleophilic aromatic substitution following the SnAr mechanism.
    Individual data points encode four process parameters for a flow reactor to run the reaction, along with the measured e-factor (defined as the ratio of the mass waste to the mass of product).
    The dataset includes 67 samples with four parameters and one objective.

    Reference:
        A.M. Schweidtmann, A.D. Clayton, N. Holmes, E. Bradford, R.A. Bourne, A.A. Lapkin. Machine learning meets continuous flow chemistry: Automated optimization towards the Pareto front of multiple objectives. Chem. Eng. J. 352 (2018) 277-282.
        [DOI](https://doi.org/10.1016/j.cej.2018.07.031).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="SnAr",
            inputs=[
                Continuous("residence_time", domain=[0.5, 2.0]),
                Continuous("ratio", domain=[1.0, 5.0]),
                Continuous("concentration", domain=[0.1, 0.5]),
                Continuous("temperature", domain=[60.0, 140.0]),
            ],
            outputs=[Continuous("impurity")],
            objectives=[Minimize("impurity")],
            data=get_data("snar.csv"),
        )


class Suzuki(Problem):
    """Suzuki reaction dataset.

    This dataset reports palladium-catalyzed Suzuki cross-coupling between 2-bromophenyltetrazole and an electron-deficient aryl boronate.
    Four reaction conditions can be controlled to maximise the reaction yield.
    The dataset includes 247 samples with four parameters and one objective.

    Reference:
        F. Häse, M. Aldeghi, R.J. Hickman, L.M. Roch, M. Christensen, E. Liles, J.E. Hein, A. Aspuru-Guzik. Olympus: a benchmarking framework for noisy optimization and experiment planning. arXiv (2020), 2010.04153.
        [DOI](https://doi.org/10.1088/2632-2153/abedc8).
        Obtained from [Olympus](https://github.com/aspuru-guzik-group/olympus).
    """

    def __init__(self):
        super().__init__(
            name="Suzuki",
            inputs=[
                Continuous("temperature", domain=[75.0, 90.0]),
                Continuous("pd_mol", domain=[0.5, 5.0]),
                Continuous("arbpin", domain=[1.0, 1.8]),
                Continuous("k3po4", domain=[1.5, 3.0]),
            ],
            outputs=[Continuous("yield")],
            objectives=[Maximize("yield")],
            data=get_data("suzuki.csv"),
        )



"""
Mixed variables single and multi-objective test problems.
"""

import numpy as np
import pandas as pd

from opti.parameter import Categorical, Continuous, Discrete
from opti.problem import Problem


class DiscreteVLMOP2(Problem):
    """VLMOP2 problem (also known as Fonzeca & Fleming), modified to contain a discrete variable.

    * 2 minimization objectives
    * 1 categorical and n continuous inputs, unconstrained

    See:
        Manson2021, MVMOO: Mixed variable multi-objective optimisation
        https://doi.org/10.1007/s10898-021-01052-9

    """

    def __init__(self, n_inputs: int = 3):
        assert n_inputs >= 2
        super().__init__(
            name="Discrete VLMOP2 test problem",
            inputs=[Categorical("x1", ["a", "b"])]
            + [Continuous(f"x{i+1}", [-2, 2]) for i in range(1, n_inputs)],
            outputs=[Continuous("y1"), Continuous("y2")],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        d = X[self.inputs.names[0]].values
        x = X[self.inputs.names[1:]].values
        n = self.n_inputs
        y1 = np.exp(-np.sum((x - n**-0.5) ** 2, axis=1))
        y2 = np.exp(-np.sum((x + n**-0.5) ** 2, axis=1))
        y1 = np.where(d == "a", 1 - y1, 1.25 - y1)
        y2 = np.where(d == "a", 1 - y2, 0.75 - y2)
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)


class DiscreteFuelInjector(Problem):
    """Fuel injector test problem, modified to contain an integer variable.

    * 4 objectives,
    * mixed variables, unconstrained

    See:
        Manson2021, MVMOO: Mixed variable multi-objective optimisation
        https://doi.org/10.1007/s10898-021-01052-9
    """

    def __init__(self):
        super().__init__(
            name="Discrete fuel injector test problem",
            inputs=[
                Discrete("x1", [0, 1, 2, 3]),
                Continuous("x2", [-2, 2]),
                Continuous("x3", [-2, 2]),
                Continuous("x4", [-2, 2]),
            ],
            outputs=[Continuous(f"y{i+1}") for i in range(4)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x1 = X["x1"].to_numpy().astype(float)
        x2 = X["x2"].to_numpy().astype(float)
        x3 = X["x3"].to_numpy().astype(float)
        x4 = X["x4"].to_numpy().astype(float)
        x1 *= 0.2
        y1 = (
            0.692
            + 0.4771 * x1
            - 0.687 * x4
            - 0.08 * x3
            - 0.065 * x2
            - 0.167 * x1**2
            - 0.0129 * x1 * x4
            + 0.0796 * x4**2
            - 0.0634 * x1 * x3
            - 0.0257 * x3 * x4
            + 0.0877 * x3**2
            - 0.0521 * x1 * x2
            + 0.00156 * x2 * x4
            + 0.00198 * x2 * x3
            + 0.0184 * x2**2
        )
        y2 = (
            0.37
            - 0.205 * x1
            + 0.0307 * x4
            + 0.108 * x3
            + 1.019 * x2
            - 0.135 * x1**2
            + 0.0141 * x1 * x4
            + 0.0998 * x4**2
            + 0.208 * x1 * x3
            - 0.0301 * x3 * x4
            - 0.226 * x3**2
            + 0.353 * x1 * x2
            - 0.0497 * x2 * x3
            - 0.423 * x2**2
            + 0.202 * x1**2 * x4
            - 0.281 * x1**2 * x3
            - 0.342 * x1 * x4**2
            - 0.245 * x3 * x4**2
            + 0.281 * x3**2 * x4
            - 0.184 * x1 * x2**2
            + 0.281 * x1 * x3 * x4
        )
        y3 = (
            0.153
            - 0.322 * x1
            + 0.396 * x4
            + 0.424 * x3
            + 0.0226 * x2
            + 0.175 * x1**2
            + 0.0185 * x1 * x4
            - 0.0701 * x4**2
            - 0.251 * x1 * x3
            + 0.179 * x3 * x4
            + 0.015 * x3**2
            + 0.0134 * x1 * x2
            + 0.0296 * x2 * x4
            + 0.0752 * x2 * x3
            + 0.0192 * x2**2
        )
        y4 = (
            0.758
            + 0.358 * x1
            - 0.807 * x4
            + 0.0925 * x3
            - 0.0468 * x2
            - 0.172 * x1**2
            + 0.0106 * x1 * x4
            + 0.0697 * x4**2
            - 0.146 * x1 * x3
            - 0.0416 * x3 * x4
            + 0.102 * x3**2
            - 0.0694 * x1 * x2
            - 0.00503 * x2 * x4
            + 0.0151 * x2 * x3
            + 0.0173 * x2**2
        )
        return pd.DataFrame({"y1": y1, "y2": y2, "y3": y3, "y4": y4})



import pprint
from typing import List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

_CAT_SEP = "§"


class Parameter:
    """Parameter base class."""

    def __init__(self, name: str, domain: Sequence, type: str = None, **kwargs):
        self.name = name
        self.domain = domain
        self.type = type
        self.extra_fields = kwargs

    def to_config(self) -> dict:
        """Return a json-serializable configuration dict."""
        conf = dict(name=self.name, type=self.type, domain=self.domain)
        conf.update(self.extra_fields)
        return conf


class Continuous(Parameter):
    """Variable that can take on any real value in the specified domain.

    Attributes:
        name (str): name of the parameter
        domain (list): [lower bound, upper bound]
    """

    def __init__(
        self,
        name: str,
        domain: Optional[Sequence] = None,
        **kwargs,
    ):
        if domain is None:
            domain = [-np.inf, np.inf]
        else:
            if len(domain) != 2:
                raise ValueError(
                    f"{name}: Domain must consist of two values [low, high]."
                )
        # convert None to +/- inf and string to float
        low = -np.inf if domain[0] is None else float(domain[0])
        high = np.inf if domain[1] is None else float(domain[1])

        if high < low:
            raise ValueError(
                f"{name}: Lower bound {low} must be less than upper bound {high}."
            )
        self.low = low
        self.high = high
        super().__init__(name=name, domain=[low, high], type="continuous", **kwargs)

    def __repr__(self):
        if np.isfinite(self.low) or np.isfinite(self.high):
            return f"Continuous('{self.name}', domain={self.domain})"
        else:
            return f"Continuous('{self.name}')"

    @property
    def bounds(self) -> Tuple[float, float]:
        """Return the domain bounds."""
        return self.low, self.high

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        return (self.low <= point) & (point <= self.high)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with values clipped to parameter bounds.
        """
        return np.clip(point, self.low, self.high)

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        low = max(self.low, np.finfo(np.float32).min)
        high = min(self.high, np.finfo(np.float32).max)
        return pd.Series(name=self.name, data=np.random.uniform(low, high, n))

    def to_config(self) -> dict:
        """Return a json-serializable configuration dict."""
        conf = dict(name=self.name, type=self.type)
        low = None if np.isinf(self.low) else float(self.low)
        high = None if np.isinf(self.high) else float(self.high)
        if low is not None or high is not None:
            conf.update({"domain": [low, high]})
        conf.update(self.extra_fields)
        return conf

    def to_unit_range(self, points):
        """Transform points to the unit range: [low, high] -> [0, 1].

        Points outside of the domain will transform to outside of [0, 1].
        Nothing is done if low == high.
        """
        if np.isclose(self.low, self.high):
            return points
        else:
            return (points - self.low) / (self.high - self.low)

    def from_unit_range(self, points):
        """Transform points from the unit range: [0, 1] -> [low, high].

        A rounding is applied to correct for numerical precision.
        Nothing is done if low == high.
        """
        if np.isclose(self.low, self.high):
            return points
        else:
            return points * (self.high - self.low) + self.low


class Discrete(Parameter):
    """Variable with a discrete domain (ordinal scale).

    Attributes:
        name (str): name of the parameter
        domain (list): list of possible numeric values
    """

    def __init__(self, name: str, domain: Sequence, **kwargs):
        if len(domain) < 1:
            raise ValueError(f"{name}: Domain must contain at least one value.")
        try:
            # convert to a sorted list of floats
            domain = np.sort(np.array(domain).astype(float)).tolist()
        except ValueError:
            raise ValueError(f"{name}: Domain contains non-numeric values.")
        if len(set(domain)) != len(domain):
            raise ValueError(f"{name}: Domain contains duplicates.")
        self.low = min(domain)
        self.high = max(domain)
        super().__init__(name, domain, type="discrete", **kwargs)

    def __repr__(self):
        return f"Discrete('{self.name}', domain={self.domain})"

    @property
    def bounds(self) -> Tuple[float, float]:
        """Return the domain bounds."""
        return self.low, self.high

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        if not np.isscalar(point):
            point = np.array(point)
        return np.isin(point, self.domain)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with values clipped to parameter bounds.
        """
        if np.isscalar(point):
            i = np.argmin(np.abs(np.array(self.domain) - point))
            return self.domain[i]
        closest = [np.argmin(np.abs(np.array(self.domain) - p)) for p in point]
        rounded = np.array(self.domain)[closest]
        if isinstance(point, np.ndarray):
            return rounded
        elif isinstance(point, pd.Series):
            return pd.Series(name=self.name, data=rounded, index=point.index)
        elif isinstance(point, pd.DataFrame):
            return pd.DataFrame({self.name: rounded}, index=point.index)

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        return pd.Series(name=self.name, data=np.random.choice(self.domain, n))

    def to_unit_range(self, points):
        """Transform points to the unit range: [low, high] -> [0, 1].

        Points outside of the domain will transform to outside of [0, 1].
        Nothing is done if low == high.
        """
        if np.isclose(self.low, self.high):
            return points
        else:
            return (points - self.low) / (self.high - self.low)

    def from_unit_range(self, points):
        """Transform points from the unit range: [0, 1] -> [low, high].

        A rounding is applied to correct for numerical precision.
        Nothing is done if low == high.
        """
        if np.isclose(self.low, self.high):
            return points
        else:
            points = points * (self.high - self.low) + self.low
            return self.round(points)


class Categorical(Parameter):
    """Variable with a categorical domain (nominal scale, values cannot be put into order).

    Attributes:
        name (str): name of the parameter
        domain (list): list possible values
    """

    def __init__(self, name: str, domain: List[str], **kwargs):
        if not isinstance(domain, list):
            raise TypeError(f"{name}: Domain must be of type list.")
        if len(domain) < 2:
            raise ValueError(f"{name}: Domain must a least contain 2 values.")
        if len(set(domain)) != len(domain):
            raise ValueError(f"{name}: Domain contains duplicates.")
        super().__init__(name, domain, type="categorical", **kwargs)

    def __repr__(self):
        return f"Categorical('{self.name}', domain={self.domain})"

    def contains(self, point):
        """Check if a point is in contained in the domain.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point` with boolean datatype.
        """
        if not np.isscalar(point):
            point = np.array(point)
        return np.isin(point, self.domain)

    def round(self, point):
        """Round a point to the closest contained values.

        Args:
            point (float, np.ndarray, pd.Series or pd.Dataframe): parameter value(s).

        Returns:
            Object of the same type as `point`.
        """
        if not np.all(self.contains(point)):
            raise ValueError(
                f"{self.name}: Cannot round values for categorical parameter."
            )
        return point

    def sample(self, n: int = 1) -> pd.Series:
        """Draw random samples from the domain."""
        return pd.Series(name=self.name, data=np.random.choice(self.domain, n))

    def to_onehot_encoding(self, points: pd.Series) -> pd.DataFrame:
        """Convert points to a one-hot encoding."""
        return pd.DataFrame(
            {f"{self.name}{_CAT_SEP}{c}": points == c for c in self.domain}, dtype=float
        )

    def from_onehot_encoding(self, points: pd.DataFrame) -> pd.Series:
        """Convert points back from one-hot encoding."""
        cat_cols = [f"{self.name}{_CAT_SEP}{c}" for c in self.domain]
        if np.any([c not in cat_cols for c in points.columns]):
            raise ValueError(
                f"{self.name}: Column names don't match categorical levels: {points.columns}, {cat_cols}."
            )
        s = points.idxmax(1).str.split(_CAT_SEP, expand=True)[1]
        s.name = self.name
        return s

    def to_dummy_encoding(self, points: pd.Series) -> pd.DataFrame:
        """Convert points to a dummy-hot encoding, dropping the first categorical level."""
        return pd.DataFrame(
            {f"{self.name}{_CAT_SEP}{c}": points == c for c in self.domain[1:]},
            dtype=float,
        )

    def from_dummy_encoding(self, points: pd.DataFrame) -> pd.Series:
        """Convert points back from dummy encoding."""
        cat_cols = [f"{self.name}{_CAT_SEP}{c}" for c in self.domain]
        if np.any([c not in cat_cols[1:] for c in points.columns]):
            raise ValueError(
                f"{self.name}: Column names don't match categorical levels: {points.columns}, {cat_cols}."
            )
        points = points.copy()
        points[cat_cols[0]] = 1 - points[cat_cols[1:]].sum(axis=1)
        s = points.idxmax(1).str.split(_CAT_SEP, expand=True)[1]
        s.name = self.name
        return s

    def to_label_encoding(self, points: pd.Series) -> pd.Series:
        """Convert points to label-encoding."""
        enc = pd.Series(range(len(self.domain)), index=list(self.domain))
        s = enc[points]
        s.index = points.index
        s.name = self.name
        return s

    def from_label_encoding(self, points: pd.Series) -> pd.Series:
        """Convert points back from label-encoding."""
        enc = np.array(self.domain)
        return pd.Series(enc[points], index=points.index)


def make_parameter(
    name: str,
    type: str,
    domain: Optional[Sequence] = None,
    **kwargs,
):
    """Make a parameter object from a configuration

    p = make_parameter(**config)

    Args:
        type (str): "continuous", "discrete" or "categorical"
        name (str): Name of the parameter
        domain (list): Domain, e.g, [0, 1], [1, 2.5, 5] or ["A", "B", "C"]
    """
    parameter = {
        "continuous": Continuous,
        "discrete": Discrete,
        "categorical": Categorical,
    }[type.lower()]
    if domain is None and parameter is not Continuous:
        raise ValueError(f"Domain not given for parameter {name}.")

    return parameter(name=name, domain=domain, **kwargs)


class Parameters:
    """Set of parameters representing either the input or the output parameter space."""

    def __init__(self, parameters):
        """
        It can be constructed either from a list / tuple (of at least one) Parameter objects
        ```
        Parameters([
            Continuous(name="foo", domain=[1, 10]),
            Discrete(name="bar", domain=[1, 2, 3, 4]),
            Categorical(name="baz", domain=["A", "B", 3]),
        ])
        ```
        or from a list / tuple of dicts
        ```
        Parameters([
            {"name": "foo", "type": "continuous", "domain": [1, 10]},
            {"name": "bar", "type": "discrete", "domain": [1, 2, 3, 4]},
            {"name": "baz", "type": "categorical", "domain": ["A", "B", 3]},
            {"name": "baz", "type": "categorical", "domain": ["A", "B", 3], extra="info"},
        ])
        ```
        In particular, Parameters().__init__ and the to_config method of each Parameter type are inverses.
        Parameters(conf).to_config() == conf
        """
        if not isinstance(parameters, (list, tuple)):
            raise TypeError("Parameters expects a list or tuple of parameters.")

        self.parameters = {}
        for d in parameters:
            if not isinstance(d, Parameter):
                d = make_parameter(**d)
            if d.name in self.parameters:
                raise ValueError(f"Duplicate parameter name {d.name}")
            self.parameters[d.name] = d

    def __repr__(self):
        return "Parameters(\n" + pprint.pformat(list(self.parameters.values())) + "\n)"

    def __iter__(self):
        return iter(self.parameters.values())

    def __getitem__(self, name):
        return self.parameters[name]

    def __len__(self):
        return len(self.parameters)

    def __add__(self, other):
        parameters = list(self.parameters.values()) + list(other.parameters.values())
        return Parameters(parameters)

    @property
    def names(self):
        return list(self.parameters.keys())

    @property
    def bounds(self) -> pd.DataFrame:
        """Return the parameter bounds."""
        for p in self:
            if isinstance(p, Categorical):
                raise TypeError(
                    f"Contains categorical parameter {p.name} which is not bounded."
                )
        return pd.DataFrame({p.name: p.bounds for p in self}, index=["min", "max"])

    def contains(self, points: pd.DataFrame) -> pd.Series:
        """Check if points are inside the domain of each parameter."""
        if isinstance(points, pd.DataFrame):
            points = points[self.names]
        b = np.stack([self[k].contains(v) for k, v in points.iteritems()], axis=1)
        return b.all(axis=1)

    def round(self, points: pd.DataFrame) -> pd.DataFrame:
        """Round points to the closest contained values."""
        return pd.concat([self[k].round(v) for k, v in points.iteritems()], axis=1)

    def sample(self, n: int = 1) -> pd.DataFrame:
        """Draw uniformly distributed random samples."""
        return pd.concat([param.sample(n) for param in self], axis=1)

    def transform(
        self,
        points: pd.DataFrame,
        continuous: str = "none",
        discrete: str = "none",
        categorical: str = "onehot-encode",
    ) -> pd.DataFrame:
        """Transfrom the given dataframe according to a set of transformation rules.

        Args:
            points (pd.DataFrame): Dataframe to transfrom. Must contain columns for each parameter and may contain additional columns.
            continuous (str, optional): Transform for continuous parameters. Options are
                - "none" (default): keep values unchanged.
                - "normalize": transforms the domain bounds to [0, 1]
            discrete (str, optional): Transform for discrete parameters. Options are
                - "none" (default): keep values unchanged.
                - "normalize": transforms the domain bounds to [0, 1]
            categorical (str, optional): Transform for categoricals. Options are
                - "onehot-encode" (default): A parameter with levels [A, B, C] transforms to 3 columns holding values [0, 1]
                - "dummy-encode": a parameter with levels [A, B, C] transforms to 2 columns holding values [0, 1]
                - "label-encode": a parameter with levels [A, B, C] transfroms to 1 columns with values [0, 1, 2]
                - "none": keep values unchanged

        Raises:
            ValueError: Unknown transform.

        Returns:
            pd.DataFrame: Transformed points. Columns that don't correspond to parameters are dropped.
        """
        transformed = []
        for p in self:
            s = points[p.name]
            if isinstance(p, Continuous):
                if continuous == "none":
                    transformed.append(s)
                elif continuous == "normalize":
                    transformed.append(p.to_unit_range(s))
                else:
                    raise ValueError(f"Unknown continuous transform {continuous}.")
            if isinstance(p, Discrete):
                if discrete == "none":
                    transformed.append(s)
                elif discrete == "normalize":
                    transformed.append(p.to_unit_range(s))
                else:
                    raise ValueError(f"Unknown discrete transform {continuous}.")
            if isinstance(p, Categorical):
                if categorical == "none":
                    transformed.append(s)
                elif categorical == "onehot-encode":
                    transformed.append(p.to_onehot_encoding(s))
                elif categorical == "dummy-encode":
                    transformed.append(p.to_dummy_encoding(s))
                elif categorical == "label-encode":
                    transformed.append(p.to_label_encoding(s))
                else:
                    raise ValueError(f"Unknown categorical transform {continuous}.")
        return pd.concat(transformed, axis=1)

    def to_config(self) -> List[dict]:
        """Configuration of the parameter space."""
        return [param.to_config() for param in self.parameters.values()]



import json
import os
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from opti.constraint import Constraint, Constraints
from opti.model import LinearModel, Model, Models
from opti.objective import Minimize, Objective, Objectives
from opti.parameter import Categorical, Continuous, Discrete, Parameter, Parameters
from opti.sampling import constrained_sampling
from opti.sampling.base import sobol_sampling

ParametersLike = Union[Parameters, List[Parameter], List[Dict]]
ObjectivesLike = Union[Objectives, List[Objective], List[Dict]]
ConstraintsLike = Union[Constraints, List[Constraint], List[Dict]]
ModelsLike = Union[Models, List[Model], List[Dict]]
DataFrameLike = Union[pd.DataFrame, Dict]
PathLike = Union[str, bytes, os.PathLike]


class Problem:
    def __init__(
        self,
        inputs: ParametersLike,
        outputs: ParametersLike,
        objectives: Optional[ObjectivesLike] = None,
        constraints: Optional[ConstraintsLike] = None,
        output_constraints: Optional[ObjectivesLike] = None,
        f: Optional[Callable] = None,
        models: Optional[ModelsLike] = None,
        data: Optional[DataFrameLike] = None,
        optima: Optional[DataFrameLike] = None,
        name: Optional[str] = None,
        **kwargs,
    ):
        """An optimization problem.

        Args:
            inputs: Input parameters.
            outputs: Output parameters.
            objectives: Optimization objectives. Defaults to minimization.
            constraints: Constraints on the inputs.
            output_constraints: Constraints on the outputs.
            f: Function to evaluate the outputs for given inputs.
                Must have the signature: f(x: pd.DataFrame) -> pd.DataFrame
            data: Experimental data.
            optima: Pareto optima.
            name: Name of the problem.
        """
        self.name = name if name is not None else "Problem"
        self.inputs = inputs if isinstance(inputs, Parameters) else Parameters(inputs)
        self.outputs = (
            outputs if isinstance(outputs, Parameters) else Parameters(outputs)
        )

        if objectives is None:
            self.objectives = Objectives([Minimize(m) for m in self.outputs.names])
        elif isinstance(objectives, Objectives):
            self.objectives = objectives
        else:
            self.objectives = Objectives(objectives)

        if isinstance(constraints, Constraints):
            pass
        elif not constraints:
            constraints = None
        else:
            constraints = Constraints(constraints)
            if len(constraints) == 0:  # no valid constraints
                constraints = None
        self.constraints = constraints

        if isinstance(output_constraints, Objectives) or output_constraints is None:
            self.output_constraints = output_constraints
        else:
            self.output_constraints = Objectives(output_constraints)

        if isinstance(models, Models) or models is None:
            self.models = models
        else:
            self.models = Models(models)

        if f is not None:
            self.f = f

        if isinstance(data, dict):
            data = pd.read_json(json.dumps(data), orient="split")

        if isinstance(optima, dict):
            optima = pd.read_json(json.dumps(optima), orient="split")

        self.set_data(data)
        self.set_optima(optima)
        self.check_problem()
        self.check_models()

    @property
    def n_inputs(self) -> int:
        return len(self.inputs)

    @property
    def n_outputs(self) -> int:
        return len(self.outputs)

    @property
    def n_objectives(self) -> int:
        return len(self.objectives)

    @property
    def n_constraints(self) -> int:
        return 0 if self.constraints is None else len(self.constraints)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        s = "Problem(\n"
        s += f"name={self.name},\n"
        s += f"inputs={self.inputs},\n"
        s += f"outputs={self.outputs},\n"
        s += f"objectives={self.objectives},\n"
        if self.output_constraints is not None:
            s += f"output_constraints={self.output_constraints},\n"
        if self.constraints is not None:
            s += f"constraints={self.constraints},\n"
        if self.models is not None:
            s += f"models={self.models},\n"
        if self.data is not None:
            s += f"data=\n{self.data.head()}\n"
        if self.optima is not None:
            s += f"optima=\n{self.optima.head()}\n"
        return s + ")"

    @staticmethod
    def from_config(config: dict) -> "Problem":
        """Create a Problem instance from a configuration dict."""
        return Problem(**config)

    def to_config(self) -> dict:
        """Return json-serializable configuration dict."""

        config = {
            "name": self.name,
            "inputs": self.inputs.to_config(),
            "outputs": self.outputs.to_config(),
            "objectives": self.objectives.to_config(),
        }
        if self.output_constraints is not None:
            config["output_constraints"] = self.output_constraints.to_config()
        if self.constraints is not None:
            config["constraints"] = self.constraints.to_config()
        if self.models is not None:
            config["models"] = self.models.to_config()
        if self.data is not None:
            config["data"] = self.data.replace({np.nan: None}).to_dict("split")
        if self.optima is not None:
            config["optima"] = self.optima.replace({np.nan: None}).to_dict("split")
        return config

    @staticmethod
    def from_json(fname: PathLike) -> "Problem":
        """Read a problem from a JSON file."""
        with open(fname, "rb") as infile:
            config = json.loads(infile.read())
        return Problem(**config)

    def to_json(self, fname: PathLike) -> None:
        """Save a problem from a JSON file."""
        with open(fname, "wb") as outfile:
            b = json.dumps(self.to_config(), ensure_ascii=False, separators=(",", ":"))
            outfile.write(b.encode("utf-8"))

    def check_problem(self) -> None:
        """Check if input and output parameters are consistent."""
        # check for duplicate names
        duplicates = set(self.inputs.names).intersection(self.outputs.names)
        if duplicates:
            raise ValueError(f"Parameter name in both inputs and outputs: {duplicates}")

        # check if all objectives refer to an output
        for obj in self.objectives:
            if obj.name not in self.outputs.names:
                raise ValueError(f"Objective refers to unknown parameter: {obj.name}")

    def check_data(self, data: pd.DataFrame) -> None:
        """Check if data is consistent with input and output parameters."""
        for p in self.inputs + self.outputs:
            # data must contain all parameters
            if p.name not in data.columns:
                raise ValueError(
                    f"Parameter {p.name} is missing. Data must contain all parameters."
                )

            # data for continuous / discrete parameters must be numeric
            if isinstance(p, (Continuous, Discrete)):
                ok = is_numeric_dtype(data[p.name]) or data[p.name].isnull().all()
                if not ok:
                    raise ValueError(
                        f"Parameter {p.name} contains non-numeric values. Data for continuous / discrete parameters must be numeric."
                    )

            # categorical levels in data must be specified
            elif isinstance(p, Categorical):
                ok = p.contains(data[p.name]) | data[p.name].isna()
                if not ok.all():
                    unknowns = data[p.name][~ok].unique().tolist()
                    raise ValueError(
                        f"Data for parameter {p.name} contains unknown values: {unknowns}. All categorical levels must be specified."
                    )

        # inputs must be complete
        for p in self.inputs:
            if data[p.name].isnull().any():
                raise ValueError(
                    f"Input parameter {p.name} has missing data. Inputs must be complete."
                )

        # outputs must have at least one observation
        for p in self.outputs:
            if data[p.name].isnull().all():
                raise ValueError(
                    f"Output parameter {p.name} has no data. Outputs must have at least one observation."
                )

    def check_models(self) -> None:
        """Check if the models are well defined"""
        if self.models is None:
            return

        for model in self.models:
            # models need to refer to output parameters
            for n in model.names:
                if n not in self.outputs.names:
                    raise ValueError(f"Model {model} refers to unknown outputs")

            if isinstance(model, LinearModel):
                if len(model.coefficients) != self.n_inputs:
                    raise ValueError(f"Model {model} has wrong number of coefficients.")

    def set_data(self, data: Optional[pd.DataFrame]) -> None:
        """Set the data."""
        if data is not None:
            for p in self.inputs:
                # Categorical levels are required to be strings. Ensure that the corresponding data is as well.
                if isinstance(p, Categorical):
                    nulls = data[p.name].isna()
                    data[p.name] = data[p.name].astype(str).mask(nulls, np.nan)

            self.check_data(data)

        self.data = data

    def get_data(self) -> pd.DataFrame:
        """Return `self.data` if it exists or an empty dataframe."""
        if self.data is None:
            return pd.DataFrame(columns=self.inputs.names + self.outputs.names)
        return self.data

    def add_data(self, data: pd.DataFrame) -> None:
        """Add a number of data points."""
        self.check_data(data)
        self.data = pd.concat([self.data, data], axis=0)

    def set_optima(self, optima: Optional[pd.DataFrame]) -> None:
        """Set the optima / Pareto front."""
        if optima is not None:
            self.check_data(optima)
        self.optima = optima

    def get_X(self, data: Optional[pd.DataFrame] = None) -> np.ndarray:
        """Return the input values in `data` or `self.data`."""
        if data is not None:
            return data[self.inputs.names].values
        return self.get_data()[self.inputs.names].values

    def get_Y(self, data: Optional[pd.DataFrame] = None) -> np.ndarray:
        """Return the output values in `data` or `self.data`."""
        if data is not None:
            return data[self.outputs.names].values
        return self.get_data()[self.outputs.names].values

    def get_XY(
        self,
        outputs: Optional[List[str]] = None,
        data: Optional[pd.DataFrame] = None,
        continuous: str = "none",
        discrete: str = "none",
        categorical: str = "none",
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return the input and output values as numeric numpy arrays.

        Rows with missing output values will be dropped.
        Input values are assumed to be complete.
        Categorical outputs are one-hot or label encoded.

        Args:
            outputs (optional): Subset of the outputs to consider.
            data (optional): Dataframe to consider instead of problem.data
        """
        if outputs is None:
            outputs = self.outputs.names
        if data is None:
            data = self.get_data()
        notna = data[outputs].notna().all(axis=1)
        X = self.inputs.transform(
            data, continuous=continuous, discrete=discrete, categorical=categorical
        )[notna].values
        Y = data[outputs][notna].values
        return X, Y

    def get_X_bounds(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return the lower and upper data bounds."""
        X = self.get_X()
        xlo = X.min(axis=0)
        xhi = X.max(axis=0)
        b = xlo == xhi
        xhi[b] = xlo[b] + 1  # prevent division by zero when dividing by (xhi - xlo)
        return xlo, xhi

    def sample_inputs(self, n_samples=10) -> pd.DataFrame:
        """Uniformly sample points from the input space subject to the constraints."""
        if self.constraints is None:
            return sobol_sampling(n_samples, self.inputs)
        return constrained_sampling(n_samples, self.inputs, self.constraints)

    def create_initial_data(self, n_samples: int = 10) -> None:
        """Create an initial data set for benchmark problems by sampling uniformly from the input space and evaluating f(x) at the sampled inputs."""
        if self.f is None:
            raise NotImplementedError("problem.f is not implemented for the problem.")
        X = self.sample_inputs(n_samples)
        Y = self.f(X)
        self.data = pd.concat([X, Y], axis=1)


def read_json(filepath: PathLike) -> Problem:
    """Read a problem specification from a JSON file."""
    return Problem.from_json(filepath)



"""
Simple 1D problems for assessing probabilistic surrogate models.
Note: these problems should be output-noisified, e.g.
```
import opti

problem = opti.problems.noisify_problem_with_gaussian(
    opti.problems.Line1D(),
    sigma=0.1
)
```
"""
import pandas as pd

from opti.parameter import Continuous, Discrete
from opti.problem import Problem

_X = pd.DataFrame({"x": [0.2, 1, 1.5, 2, 3, 3.1, 6.5, 7, 7.2, 7.5]})


class Line1D(Problem):
    """A line."""

    def __init__(self):
        def f(X: pd.DataFrame) -> pd.DataFrame:
            return pd.DataFrame({"y": X.eval("0.1 * x + 1")}, index=X.index)

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=pd.concat([_X, f(_X)], axis=1),
        )


class Parabola1D(Problem):
    """A parabola."""

    def __init__(self):
        def f(X: pd.DataFrame) -> pd.DataFrame:
            return pd.DataFrame(
                {"y": X.eval("0.025 * (x - 5) ** 2 + 1")}, index=X.index
            )

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=pd.concat([_X, f(_X)], axis=1),
        )


class Sinus1D(Problem):
    """A sinus-function with one full period over the domain."""

    def __init__(self):
        def f(X: pd.DataFrame) -> pd.DataFrame:
            return pd.DataFrame(
                {"y": X.eval("sin(x * 2 * 3.14159 / 10) / 2 + 2")}, index=X.index
            )

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=pd.concat([_X, f(_X)], axis=1),
        )


class Sigmoid1D(Problem):
    """A smooth step at x=5."""

    def __init__(self):
        def f(X: pd.DataFrame) -> pd.DataFrame:
            return pd.DataFrame(
                {"y": X.eval("1 / (1 + exp(-2 * (x - 5))) + 1")}, index=X.index
            )

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Continuous("y", [0, 3])],
            f=f,
            data=pd.concat([_X, f(_X)], axis=1),
        )


class Step1D(Problem):
    """A discrete step at x=1.1."""

    def __init__(self):
        def f(X: pd.DataFrame) -> pd.DataFrame:
            return pd.DataFrame({"y": X.eval("x > 1.1").astype(float)}, index=X.index)

        super().__init__(
            inputs=[Continuous("x", [0, 10])],
            outputs=[Discrete("y", [0, 1])],
            f=f,
            data=pd.concat([_X, f(_X)], axis=1),
        )



# flake8: noqa
import opti.sampling.polytope
import opti.sampling.simplex
import opti.sampling.sphere
from opti.sampling.base import constrained_sampling, rejection_sampling, sobol_sampling
from opti.sampling.polytope import polytope_sampling



"""
This module provides functions to uniformly sample points subject to a system of linear
inequality constraints, :math:`Ax <= b` (convex polytope), and linear equality
constraints, :math:`Ax = b` (affine projection).

A comparison of MCMC algorithms to generate uniform samples over a convex polytope is
given in [Chen2018]_. Here, we use the Hit & Run algorithm described in [Smith1984]_.
The R-package `hitandrun`_ provides similar functionality to this module.

References
----------
.. [Chen2018] Chen Y., Dwivedi, R., Wainwright, M., Yu B. (2018) Fast MCMC Sampling
    Algorithms on Polytopes. JMLR, 19(55):1−86
    https://arxiv.org/abs/1710.08165
.. [Smith1984] Smith, R. (1984). Efficient Monte Carlo Procedures for Generating
    Points Uniformly Distributed Over Bounded Regions. Operations Research,
    32(6), 1296-1308.
    www.jstor.org/stable/170949
.. _`hitandrun`: https://cran.r-project.org/web/packages/hitandrun/index.html
"""
from typing import Generator, Tuple

import numpy as np
import pandas as pd
import scipy.linalg
import scipy.optimize

from opti.constraint import Constraints, LinearEquality, LinearInequality
from opti.parameter import Parameters


def _chebyshev_center(A: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Find the center of the convex polytope Ax <= b."""
    res = scipy.optimize.linprog(
        np.r_[np.zeros(A.shape[1]), -1],
        A_ub=np.hstack([A, np.linalg.norm(A, axis=1, keepdims=True)]),
        b_ub=b,
        bounds=(None, None),
    )
    if not res.success:
        raise Exception("Unable to find Chebyshev center")
    return res.x[:-1]


def _affine_subspace(A: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Compute a orthonormal basis of the nullspace of A, and a solution to Ax = b.
    This allows to to construct arbitrary solutions as the sum of any vector in the
    nullspace, plus the particular solution.
    """
    N = scipy.linalg.null_space(A)
    xp = np.linalg.pinv(A) @ b
    return N, xp


def _hitandrun(
    A: np.ndarray, b: np.ndarray, x0: np.ndarray
) -> Generator[np.ndarray, None, None]:
    """Generator for uniform sampling from the convex polytope Ax <= b using the
    Hit & Run algorithm described in [Smith1984].
    """
    assert A.shape[1] == len(x0)

    x = x0

    while True:
        # sample random direction from unit hypersphere
        direction = np.random.randn(A.shape[1])
        direction /= np.linalg.norm(direction)

        # distances to each face from the current point in the sampled direction
        D = (b - x @ A.T) / (direction @ A.T)

        # distance to the closest face in and opposite to direction
        lo = np.max(D[D < 0])
        hi = np.min(D[D > 0])

        # make random step
        x += np.random.uniform(lo, hi) * direction
        yield x


def _get_AbNx(parameters, constraints):
    """Get a matrix representation of a system of linear constraints and box bounds.

    Convert all linear constraints and box bounds to matrix form:
        A_eq x = b_eq
        A_ineq x <= b_ineq
    compute the nullspace N and a particular solution xp for A_eq x = b_eq and
    transform the linear constraints to the subspace given by N, so that
        At xt <= bt
        x = xt N.T + xp
    """
    nD = len(parameters)
    nC = len(constraints)

    # matrix form A, b of all linear constraints, A[is_eq] are the linear equality constraints
    A = np.empty((nC, nD))
    b = np.empty(nC)
    is_eq = np.zeros(nC, dtype=bool)

    for i, c in enumerate(constraints):
        lhs = np.zeros(len(parameters))
        for ni, li in zip(c.names, c.lhs):
            lhs[parameters.names.index(ni)] = li
        A[i] = lhs
        b[i] = c.rhs
        is_eq[i] = c.is_equality

    # compute null space and particular solution corresponding to the linear equalities
    if sum(is_eq) > 0:
        N, xp = _affine_subspace(A[is_eq], b[is_eq])
    else:
        N = np.eye(nD)
        xp = np.zeros(nD)

    # add the inequality constraints corresponding to the box bounds
    lower = parameters.bounds.loc["min"]
    upper = parameters.bounds.loc["max"]
    A_ineq = np.row_stack([-np.eye(nD), np.eye(nD)])
    b_ineq = np.r_[-np.array(lower), np.array(upper)]

    if sum(~is_eq) > 0:
        A_ineq = np.r_[A_ineq, A[~is_eq]]
        b_ineq = np.r_[b_ineq, b[~is_eq]]

    # project to the affine subspace given by the equality constraints
    At = A_ineq @ N
    bt = b_ineq - A_ineq @ xp

    return At, bt, N, xp


def polytope_sampling(
    n_samples: int, parameters: Parameters, constraints: Constraints, thin: int = 100
) -> pd.DataFrame:
    """Hit-and-run method to sample uniformly under linear constraints.

    Args:
        n_samples (int): Number of samples.
        parameters (opti.Parameters): Parameter space.
        constraints (opti.Constraints): Constraints on the parameters.
        thin (int, optional): Thinning factor of the generated samples.

    Returns:
        array, shape=(n_samples, dimension): Randomly sampled points.
    """
    for c in constraints:
        if not isinstance(c, (LinearEquality, LinearInequality)):
            raise Exception("Polytope sampling only works for linear constraints.")

    At, bt, N, xp = _get_AbNx(parameters, constraints)

    # hit & run sampling
    x0 = _chebyshev_center(At, bt)
    sampler = _hitandrun(At, bt, x0)
    X = np.empty((n_samples, At.shape[1]))
    for i in range(n_samples):
        for _ in range(thin - 1):
            next(sampler)
        X[i] = next(sampler)

    # project back
    X = X @ N.T + xp
    return pd.DataFrame(columns=parameters.names, data=X)



import numpy as np
import pandas as pd
from scipy import stats

from opti.problems import ZDT2, Cake
from opti.tools.noisify import _add_noise_to_data, noisify_problem_with_gaussian


def test_add_noise_to_data():
    cake = Cake()
    cake.f = lambda df: pd.DataFrame(
        data=np.zeros((len(df), cake.n_outputs)),
        columns=cake.outputs.names,
    )

    def no_noise(Y):
        return Y

    def to_zero(Y):
        return Y * 0

    Y = cake.get_data()[cake.outputs.names]
    Y_not_noisy = _add_noise_to_data(Y, [no_noise] * len(Y), cake.outputs)
    assert np.allclose(Y, Y_not_noisy)

    # outputs are clipped to their domain bounds after noise is applied, which is why the first output ("calories") is set to 300.
    Y_not_noisy = _add_noise_to_data(Y, [to_zero] * len(Y), cake.outputs)
    assert np.allclose(Y_not_noisy, [300, 0, 0])


def test_noisify_problem_with_gaussian():
    n_samples = 5000
    mu = 0.1
    sigma = 0.05

    zdt2 = ZDT2()
    zdt2.create_initial_data(n_samples)
    zdt2_gaussian = noisify_problem_with_gaussian(zdt2, mu=mu, sigma=sigma)

    reference = np.clip(stats.norm.rvs(loc=mu, scale=sigma, size=n_samples), 0, 1)
    Y_noise = zdt2_gaussian.get_Y() - zdt2.get_Y()
    for col in Y_noise.T:
        s, _ = stats.ks_2samp(col, reference)
        assert s < 0.1



from typing import Callable, List

import pandas as pd
from scipy.stats import norm

from opti.parameter import Parameters
from opti.problem import Problem


def _add_noise_to_data(
    df: pd.DataFrame, noisifiers: List[Callable], outputs: Parameters
) -> pd.DataFrame:
    return pd.concat(
        [
            output.round(noisify(df[output.name]))
            for output, noisify in zip(outputs, noisifiers)
        ],
        axis=1,
    )


def noisify_problem(
    problem: Problem,
    noisifiers: List[Callable],
) -> Problem:
    """Creates a new problem that is based on the given one plus noise on the outputs.

    Args:
        problem: given problem where we will add noise
        noisifiers: list of functions that add noise to the outputs

    Returns: new problem with noise on the output
    """

    def noisy_f(X):
        return _add_noise_to_data(problem.f(X), noisifiers, problem.outputs)

    if problem.data is not None:
        data = problem.get_data()
        X = data[problem.inputs.names]
        Yn = _add_noise_to_data(
            data[problem.outputs.names], noisifiers, problem.outputs
        )
        data = pd.concat([X, Yn], axis=1)
    else:
        data = None

    return Problem(
        inputs=problem.inputs,
        outputs=problem.outputs,
        objectives=problem.objectives,
        constraints=problem.constraints,
        output_constraints=problem.output_constraints,
        f=noisy_f,
        data=data,
        name=problem.name,
    )


def noisify_problem_with_gaussian(problem: Problem, mu: float = 0, sigma: float = 0.05):
    """
    Given an instance of a problem, this returns the problem with additive Gaussian noise
    Args:
        problem: problem instance where we add the noise
        mu: mean of the Gaussian noise to be added
        sigma: standard deviation of the Gaussian noise to be added

    Returns: input problem with additive Gaussian noise
    """

    def noisify(y):
        rv = norm(loc=mu, scale=sigma)
        return y + rv.rvs(len(y))

    return noisify_problem(problem, noisifiers=[noisify] * len(problem.outputs))



# flake8: noqa
from opti.problems.datasets import (
    HPLC,
    Alkox,
    BaumgartnerAniline,
    BaumgartnerBenzamide,
    Benzylation,
    Cake,
    Fullerenes,
    Photodegradation,
    ReizmanSuzuki,
    SnAr,
    Suzuki,
)
from opti.problems.detergent import (
    Detergent,
    Detergent_NChooseKConstraint,
    Detergent_OutputConstraint,
    Detergent_TwoOutputConstraints,
)
from opti.problems.mixed import DiscreteFuelInjector, DiscreteVLMOP2
from opti.problems.multi import (
    Daechert1,
    Daechert2,
    Daechert3,
    Hyperellipsoid,
    OmniTest,
    Poloni,
    Qapi1,
    WeldedBeam,
)
from opti.problems.single import (
    Ackley,
    Branin,
    Himmelblau,
    Michalewicz,
    Rastrigin,
    Rosenbrock,
    Schwefel,
    Sphere,
    ThreeHumpCamel,
    Zakharov,
    Zakharov_Categorical,
    Zakharov_Constrained,
    Zakharov_NChooseKConstraint,
)
from opti.problems.univariate import Line1D, Parabola1D, Sigmoid1D, Sinus1D, Step1D
from opti.problems.zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6



import json

import numpy as np
import pandas as pd
import pytest

import opti
from opti.model import LinearModel
from opti.objective import Minimize
from opti.parameter import Continuous, Parameters
from opti.problem import Problem


def test_parameters():
    # set inputs & outputs as Parameters object
    problem = opti.Problem(
        inputs=Parameters([opti.Continuous("x")]),
        outputs=Parameters([opti.Continuous("y")]),
    )
    assert isinstance(problem.inputs, opti.Parameters)
    assert isinstance(problem.outputs, opti.Parameters)

    # set inputs & outputs as a list of Parameter
    problem = opti.Problem(
        inputs=[opti.Continuous("x", domain=[0, 1])],
        outputs=[opti.Continuous("y", domain=[0, 1])],
    )
    assert isinstance(problem.inputs, opti.Parameters)
    assert isinstance(problem.outputs, opti.Parameters)

    # set inputs & outputs as list of dicts
    problem = opti.Problem(
        inputs=[
            {"name": "x1", "type": "continuous", "domain": [1, 10]},
            {"name": "x2", "type": "categorical", "domain": ["A", "B", 3]},
        ],
        outputs=[
            {"name": "y1", "type": "continuous", "domain": [1, 10]},
            {"name": "y2", "type": "discrete", "domain": [1, 2, 3]},
        ],
    )
    assert isinstance(problem.inputs, opti.Parameters)
    assert isinstance(problem.outputs, opti.Parameters)


def test_properties():
    problem = opti.Problem(
        inputs=[opti.Continuous(f"x{i}") for i in range(10)],
        outputs=[opti.Continuous(f"y{i}") for i in range(4)],
        objectives=[opti.Minimize(f"y{i}") for i in range(2)],
    )
    assert problem.n_inputs == 10
    assert problem.n_outputs == 4
    assert problem.n_objectives == 2
    assert problem.n_constraints == 0

    problem = opti.Problem(
        inputs=[opti.Continuous(f"x{i}") for i in range(10)],
        outputs=[opti.Continuous(f"y{i}") for i in range(4)],
        constraints=[opti.LinearEquality([f"x{i}" for i in range(10)], rhs=1)],
    )
    assert problem.n_inputs == 10
    assert problem.n_outputs == 4
    assert problem.n_objectives == 4
    assert problem.n_constraints == 1


def test_json(tmpdir):
    # test loading from json
    problem = opti.read_json("examples/bread.json")
    assert len(problem.inputs) == 11
    assert len(problem.outputs) == 4
    assert problem.constraints is not None
    assert problem.name is not None

    data = pd.DataFrame(
        columns=problem.inputs.names + problem.outputs.names,
        data=np.zeros((3, len(problem.inputs) + len(problem.outputs))),
    )
    data["yeast_type"] = "A"
    problem.set_data(data)

    # objectives are optional and should be set to minimize if not given
    config = json.load(open("examples/bread.json"))
    config.pop("objectives")
    problem = opti.Problem.from_config(config)
    assert len(problem.objectives) == len(problem.outputs)
    for obj in problem.objectives:
        assert isinstance(obj, opti.objective.Minimize)

    # constraints are optional
    config = json.load(open("examples/bread.json"))
    config.pop("constraints")
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None

    # serialize to json
    tmpfile = tmpdir.join("foo.json")
    problem = opti.Problem.from_json("examples/bread.json")
    problem.to_json(tmpfile)

    # json compliance with respect to Inf and Nan
    problem = opti.problems.Ackley()
    problem.outputs["y"].low = 0  # output domain is now [0, Inf]
    problem.to_json(tmpfile)

    conf = json.load(open(tmpfile))
    assert conf["outputs"][0]["domain"][1] is None

    problem = opti.Problem.from_json(tmpfile)
    assert problem.outputs.bounds.loc["min", "y"] == 0
    assert problem.outputs.bounds.loc["max", "y"] == np.inf

    # handling of unicode characters
    problem = opti.Problem(
        inputs=[opti.Continuous("öäüßèé", [0, 1])],
        outputs=[opti.Continuous("/[].#@²³$")],
    )
    problem.to_json(tmpfile)
    problem2 = opti.read_json(tmpfile)
    assert np.all(problem2.inputs.names == problem.inputs.names)
    assert np.all(problem2.outputs.names == problem.outputs.names)

    # parse values scientific notation
    problem = opti.read_json("examples/test_scientific_notation.json")
    assert np.allclose(problem.data["x"], [0.9, 1.7e12])


def test_inconsistent_parameters():
    # duplicate parameter name in inputs and outputs
    with pytest.raises(ValueError):
        opti.Problem(inputs=[opti.Continuous("x")], outputs=[opti.Continuous("x")])


def test_inconsistent_objectives():
    # parameter name in objective not among inputs & outputs
    with pytest.raises(ValueError):
        opti.Problem(
            inputs=[opti.Continuous("x")],
            outputs=[opti.Continuous("y")],
            objectives=[opti.objective.Minimize("z")],
        )


def test_set_data():
    # test with continuous and discrete parameters
    problem = opti.Problem(
        inputs=[opti.Discrete("x", [0, 1, 2])], outputs=[opti.Continuous("y")]
    )

    # this works
    problem.set_data(data=pd.DataFrame({"x": [0, 1], "y": [0, 1]}))

    # missing parameter in data
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"y": [0, 1, 2]}))
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1, 2]}))

    # missing value in inputs
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1, None], "y": [0, 1, 2]}))

    # missing output values in outputs are ok
    problem.set_data(data=pd.DataFrame({"x": [0, 1, 2], "y": [0, 1, None]}))

    # outputs with all missing values are not ok
    with pytest.raises(ValueError):
        problem.set_data(data=pd.DataFrame({"x": [0, 1], "y": [None, None]}))

    # non-numeric value for continuous / discrete parameter
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": ["oops", 1, 2], "y": [0, 1, 2]}))
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": [0, 1, 2], "y": ["oops", 1, 2]}))

    # test with categorical parameter
    problem = opti.Problem(
        inputs=[opti.Categorical("x", ["A", "B"])], outputs=[opti.Continuous("y")]
    )

    # this works
    problem.set_data(data=pd.DataFrame({"x": ["A", "B", "A"], "y": [0, 1, 0.5]}))

    # categorical levels in data must be known
    with pytest.raises(ValueError):
        problem.set_data(pd.DataFrame({"x": ["A", "B", "C", None], "y": [0, 1, 2, 3]}))

    # test the case where a data frame with numeric values is passed for a categorical parameter
    problem = opti.Problem(
        inputs=[{"name": "x", "type": "categorical", "domain": ["0", "1", "2"]}],
        outputs=[{"name": "y", "type": "continuous", "domain": [0, 1]}],
    )
    problem.set_data(pd.DataFrame({"x": [0, 1, 2], "y": [0.1, 0.7, 0.3]}))


def test_add_data():
    problem = opti.Problem(
        inputs=[opti.Continuous("x")], outputs=[opti.Continuous("y")]
    )

    # adding data when no data exists
    problem.add_data(data=pd.DataFrame({"x": [0, 1], "y": [0, 1]}))
    assert len(problem.data) == 2

    # adding new data with extra column
    problem.add_data(data=pd.DataFrame({"x": [2], "y": [2], "z": [2]}))
    assert len(problem.data) == 3
    assert "z" in problem.data.columns

    # adding data with missing input value
    with pytest.raises(ValueError):
        problem.add_data(data=pd.DataFrame({"x": [np.nan], "y": [3]}))

    # adding data with missing input value
    with pytest.raises(ValueError):
        problem.add_data(data=pd.DataFrame({"x": [3]}))


def test_get_XY():
    problem = opti.Problem(
        inputs=[
            opti.Continuous("x1", [0, 1]),
            opti.Discrete("x2", [0, 1, 5]),
            opti.Categorical("x3", ["A", "B", "C"]),
        ],
        outputs=[opti.Continuous("y1"), opti.Continuous("y2")],
        data=pd.DataFrame(
            {
                "x1": [0.2, 0.5, 0.6],
                "x2": [0, 1, 5],
                "x3": ["A", "C", "B"],
                "y1": [1, 0.8, 0.4],
                "y2": [0, 0.2, np.nan],
            }
        ),
    )

    X = problem.get_X()
    assert X.shape == (3, 3)
    Y = problem.get_Y()
    assert Y.shape == (3, 2)

    # nans are dropped
    X, Y = problem.get_XY()
    assert X.shape == (2, 3)
    assert Y.shape == (2, 2)

    # using transforms
    X, Y = problem.get_XY(
        continuous="normalize", discrete="normalize", categorical="onehot-encode"
    )
    assert X.shape == (2, 5)
    assert Y.shape == (2, 2)

    # consider subset of outputs
    X, Y = problem.get_XY(outputs=["y1"])
    assert X.shape == (3, 3)
    assert Y.shape == (3, 1)

    # pass data
    data = pd.DataFrame(
        {
            "x1": [0, 0, 0],
            "x2": [0, 0, 0],
            "x3": ["A", "B", "C"],
            "y1": [1, np.nan, 0.4],
            "y2": [0, 0.2, np.nan],
        }
    )
    X, Y = problem.get_XY(data=data)
    assert X.shape == (1, 3)
    assert Y.shape == (1, 2)


def test_X_bounds():
    # Obtaining input data bounds should work for continuous and discrete parameters.
    # For parameters with 0 variation, require xhi = xlo + 1.
    problem = opti.Problem(
        inputs=[
            opti.Continuous("x1", [-1, 1]),
            opti.Continuous("x2", [-1, 1]),
            opti.Discrete("x3", [0, 1, 2]),
        ],
        outputs=[opti.Continuous("y")],
        data=pd.DataFrame({"x1": [-0.2, 0.5], "x2": [0, 0], "x3": [2, 0], "y": [0, 1]}),
    )
    xlo, xhi = problem.get_X_bounds()
    assert np.allclose(xlo, [-0.2, 0, 0])
    assert np.allclose(xhi, [0.5, 1, 2])


def test_empty_constraints():
    # test handling of empty constraints
    problem = opti.Problem(
        inputs=[opti.Continuous("x")],
        outputs=[opti.Continuous("y")],
        constraints=[],
    )
    assert problem.constraints is None

    config = {
        "inputs": [{"type": "continuous", "name": "x", "domain": [0, 1]}],
        "outputs": [{"type": "continuous", "name": "y", "domain": [0, 1]}],
        "constraints": [],
    }
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None

    config = {
        "inputs": [{"type": "continuous", "name": "x"}],
        "outputs": [{"type": "continuous", "name": "y"}],
        "constraints": [
            {"names": [], "lhs": [], "rhs": 1, "type": "linear-inequality"}
        ],
    }
    problem = opti.Problem.from_config(config)
    assert problem.constraints is None


def test_config():
    # test if the configuration dict is strictly json-compliant
    problem = opti.problems.Ackley()
    problem.create_initial_data(2)
    problem.data.loc[0, "y"] = np.nan
    problem.outputs["y"].low = 0  # output domain is now [0, Inf]

    conf = problem.to_config()
    # the upper bound np.inf is converted to None
    assert conf["outputs"][0]["domain"][1] is None
    # the datapoint y = np.nan is converted to None
    assert conf["data"]["data"][0][-1] is None


def test_optima():
    problem = opti.read_json("examples/simple.json")
    assert len(problem.optima) == 2


def test_models():
    D = 5
    M = 3
    problem = Problem(
        inputs=[Continuous(f"x{i}", [0, 1]) for i in range(D)],
        outputs=[Continuous(f"y{i}") for i in range(M)],
        objectives=[Minimize(f"y{i}") for i in range(M - 1)],
        models=[LinearModel([f"y{i}"], coefficients=np.ones(D)) for i in range(M)],
    )

    X = problem.sample_inputs(10)
    Y = problem.models(X)
    Z = problem.objectives(Y)
    assert list(Y.columns) == problem.outputs.names
    assert list(Z.columns) == problem.objectives.names



from itertools import product
from typing import Optional, Union

import numpy as np
import pandas as pd

import opti
from opti.constraint import LinearInequality, NonlinearInequality
from opti.parameter import Continuous
from opti.problem import Problem


class Qapi1(Problem):
    """Constrained problem from the Qriteria API tests.
    Note that while the function is convex, the constraints are not.

    minimize
        f1(x) = (x1 - 2)^2 + (x2 - 1)^2
        f2(x) = x1^2 + (x2 - 3)^2
    for
        x1 in [0, inf)
        x2 in (-inf, inf)
    s.t.
        0 <= x1
        c1(x) = - x1^2 + x2 <= 0
        c2(x) = - x1 - x2 + 2 <= 0

    The ideal point is [0, 0] and the nadir is [8, 8].
    """

    def __init__(self):
        super().__init__(
            name="Constrained bi-objective problem",
            inputs=[Continuous("x1", [0, 10]), Continuous("x2", [-10, 10])],
            outputs=[Continuous("y1"), Continuous("y2")],
            constraints=[
                NonlinearInequality("x2 - x1**2"),
                NonlinearInequality("2 - x1 - x2"),
            ],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "y1": X.eval("(x1 - 2)**2 + (x2 - 1)**2"),
                "y2": X.eval("x1**2 + (x2 - 3)**2"),
            }
        )


class Hyperellipsoid(Problem):
    """Hyperellipsoid in n dimensions

    minimize
        f_m(x) = x_m    m = 1, ... n
    for
        x in R^n
    s.t.
        sum((x / a)^2) - 1 <= 0

    The ideal point is -a and the is nadir 0^n.

    Args:
        n (int, optional): Dimension of the hyperellipsoid. Defaults to 5.
        a (list-like, optional): Half length of principal axes. a = None or a = [1, ...] results in a hypersphere.
    """

    def __init__(self, n: int = 5, a: Optional[Union[list, np.ndarray]] = None):
        if a is None:
            a = np.ones(n)
            constr = " + ".join([f"x{i+1}**2" for i in range(n)]) + " - 1"
        else:
            a = np.array(a).squeeze()
            if len(a) != n:
                raise ValueError("Dimension of half axes doesn't match input dimension")
            constr = " + ".join([f"(x{i+1}/{a[i]})**2" for i in range(n)]) + " - 1"
        self.a = a

        super().__init__(
            name="Hyperellipsoid",
            inputs=[Continuous(f"x{i+1}", [-a[i], a[i]]) for i in range(n)],
            outputs=[Continuous(f"y{i+1}", [-a[i], a[i]]) for i in range(n)],
            constraints=[NonlinearInequality(constr)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        y = X[self.inputs.names]
        y.columns = self.outputs.names
        return y

    def get_optima(self, n=10) -> pd.DataFrame:
        X = opti.sampling.sphere.sample(self.n_inputs, n, positive=True)
        X = np.concatenate([-np.eye(self.n_inputs), -X], axis=0)[:n]
        Y = self.a * X
        return pd.DataFrame(
            data=np.column_stack([X, Y]),
            columns=self.inputs.names + self.outputs.names,
        )


class Daechert1(Problem):
    """Problem with a non-convex Pareto front.

    From Dächert & Teichert 2020, An improved hyperboxing algorithm for calculating a Pareto front representation, https://arxiv.org/abs/2003.14249

    The ideal point is [-1.37, -1.61, -4] and the nadir is [0, 0, -1.44].
    """

    def __init__(self):
        super().__init__(
            name="Daechert-1",
            inputs=[
                Continuous("x1", domain=[0, np.pi]),
                Continuous("x2", domain=[0, 10]),
                Continuous("x3", domain=[1.2, 10]),
            ],
            outputs=[Continuous(f"y{i+1}") for i in range(3)],
            constraints=[NonlinearInequality("- cos(x1) - exp(-x2) + x3")],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame({"y1": -X["x1"], "y2": -X["x2"], "y3": -X["x3"] ** 2})


class Daechert2(Problem):
    """Unconstrained problem with a Pareto front resembling a comet.

    From Dächert & Teichert 2020, An improved hyperboxing algorithm for calculating a Pareto front representation, https://arxiv.org/abs/2003.14249

    minimize
        f1(x) = (1 + x3) (x1^3 x2^2 - 10 x1 - 4 x2)
        f2(x) = (1 + x3) (x1^3 x2^2 - 10 x1 + 4 x2)
        f3(x) = 3 (1 + x3) x1^2
    s.t.
         1 <= x1 <= 3.5
        -2 <= x2 <= 2
         0 <= x3 <= 1

    The ideal point is [-70.19, -70.19, 3] and the nadir is [4, 4, 73.5].
    """

    def __init__(self):
        super().__init__(
            name="Daechert-2",
            inputs=[
                Continuous("x1", domain=[1, 3.5]),
                Continuous("x2", domain=[-2, 2]),
                Continuous("x3", domain=[0, 1]),
            ],
            outputs=[Continuous(f"y{i + 1}") for i in range(3)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "y1": X.eval("(1 + x3) * (x1**3 * x2**2 - 10 * x1 - 4 * x2)"),
                "y2": X.eval("(1 + x3) * (x1**3 * x2**2 - 10 * x1 + 4 * x2)"),
                "y3": X.eval("3 * (1 + x3) * x1**2"),
            }
        )


class Daechert3(Problem):
    """Modification of DTLZ7, with a Pareto consisting of 4 disconnected parts.

    From Dächert & Teichert 2020, An improved hyperboxing algorithm for calculating a Pareto front representation, https://arxiv.org/abs/2003.14249

    The ideal point is [0, 0, 2.61] and the nadir is [0.86, 0.86, 6].
    """

    def __init__(self):
        super().__init__(
            name="Daechert-3",
            inputs=[Continuous(f"x{i+1}", domain=[0, 1]) for i in range(2)],
            outputs=[Continuous(f"y{i+1}") for i in range(3)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names]
        return pd.DataFrame(
            {
                "y1": X["x1"],
                "y2": X["x2"],
                "y3": 6 - np.sum(x * (1 + np.sin(3 * np.pi * x)), axis=1),
            }
        )


class OmniTest(Problem):
    """Bi-objective benchmark problem with D inputs and a multi-modal Pareto set.

    It has 3^D Pareto subsets in the decision space corresponding to the same Pareto front.

    Reference:
        Deb & Tiwari "Omni-optimizer: A generic evolutionary algorithm for single and multi-objective optimization"
    """

    def __init__(self, n_inputs: int = 2):
        super().__init__(
            name="Omni",
            inputs=[Continuous(f"x{i+1}", domain=[0, 6]) for i in range(2)],
            outputs=[Continuous("y1"), Continuous("y2")],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X[self.inputs.names]
        return pd.DataFrame(
            {
                "y1": np.sum(np.sin(np.pi * X), axis=1),
                "y2": np.sum(np.cos(np.pi * X), axis=1),
            }
        )

    def get_optima(self) -> pd.DataFrame:
        n = 11  # points per set (3^D sets)
        s = [np.linspace(1, 1.5, n) + 2 * i for i in range(3)]
        C = list(
            product(
                *[
                    s,
                ]
                * self.n_inputs
            )
        )
        C = np.moveaxis(C, 1, 2).reshape(-1, 2)
        X = pd.DataFrame(C, columns=self.inputs.names)
        XY = pd.concat([X, self.f(X)], axis=1)
        XY["_patch"] = np.repeat(np.arange(3**self.n_inputs), n)
        return XY


class Poloni(Problem):
    """Poloni benchmark problem."""

    def __init__(self):
        super().__init__(
            name="Poloni function",
            inputs=[Continuous(f"x{i+1}", [-np.pi, np.pi]) for i in range(2)],
            outputs=[Continuous("y1"), Continuous("y2")],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x1, x2 = self.get_X(X).T
        A1 = 0.5 * np.sin(1) - 2 * np.cos(1) + np.sin(2) - 1.5 * np.cos(2)
        A2 = 1.5 * np.sin(1) - np.cos(1) + 2 * np.sin(2) - 0.5 * np.cos(2)
        B1 = 0.5 * np.sin(x1) - 2 * np.cos(x1) + np.sin(x2) - 1.5 * np.cos(x2)
        B2 = 1.5 * np.sin(x1) - np.cos(x1) + 2 * np.sin(x2) - 0.5 * np.cos(x2)
        return pd.DataFrame(
            {
                "y1": 1 + (A1 - B1) ** 2 + (A2 - B2) ** 2,
                "y2": (x1 + 3) ** 2 + (x2 + 1) ** 2,
            },
            index=X.index,
        )


class WeldedBeam(Problem):
    """Design optimization of a welded beam.

    This is a bi-objective problem with 4 inputs and 3 (non-)linear inequality constraints.
    The two objectives are the fabrication cost of the beam and the deflection of the end of the beam under the applied load P.
    The load P is fixed at 6000 lbs, and the distance L is fixed at 14 inch.

    Note that for simplicity the constraint shear stress < 13600 psi is not included.

    See https://www.mathworks.com/help/gads/multiobjective-optimization-welded-beam.html
    """

    def __init__(self):
        super().__init__(
            name="Welded beam problem",
            inputs=[
                Continuous("h", [0.125, 5]),  # thickness of welds
                Continuous("l", [0.1, 10]),  # length of welds
                Continuous("t", [0.1, 10]),  # height of beam
                Continuous("b", [0.125, 5]),  # width of beam
            ],
            outputs=[Continuous("cost"), Continuous("deflection")],
            constraints=[
                # h <= b, weld thickness cannot exceed beam width
                LinearInequality(["h", "b"], lhs=[1, -1], rhs=0),
                # normal stress on the welds cannot exceed 30000 psi
                NonlinearInequality("6000 * 6 * 14 / b / t**3 - 30000"),
                # buckling load capacity must exceed 6000 lbs
                NonlinearInequality(
                    "6000 - 60746.022 * (1 - 0.0282346 * t) * t * b**4"
                ),
            ],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x1, x2, x3, x4 = self.get_X(X).T
        return pd.DataFrame(
            {
                "cost": 1.10471 * x1**2 * x2 + 0.04811 * x3 * x4 * (14 + x2),
                "deflection": 2.1952 / (x4 * x3**3),
            },
            index=X.index,
        )



"""
ZDT benchmark problem suite.
All problems are bi-objective, have D continuous inputs and are unconstrained.

Zitzler, Deb, Thiele 2000 - Comparison of Multiobjective Evolutionary Algorithms: Empirical Results
http://dx.doi.org/10.1162/106365600568202
"""
import numpy as np
import pandas as pd

from opti.parameter import Continuous
from opti.problem import Problem


class ZDT1(Problem):
    """ZDT-1 benchmark problem."""

    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT-1 problem",
            inputs=[Continuous(f"x{i+1}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i+1}", [0, np.inf]) for i in range(2)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names[1:]].to_numpy()
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x, axis=1)
        y1 = X["x1"].to_numpy()
        y2 = g * (1 - (y1 / g) ** 0.5)
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.sqrt(x)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT2(Problem):
    """ZDT-2 benchmark problem."""

    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT-2 problem",
            inputs=[Continuous(f"x{i+1}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i+1}", [0, np.inf]) for i in range(2)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names[1:]].to_numpy()
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x, axis=1)
        y1 = X["x1"].to_numpy()
        y2 = g * (1 - (y1 / g) ** 2)
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.power(x, 2)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT3(Problem):
    """ZDT-3 benchmark problem."""

    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT-3 problem",
            inputs=[Continuous(f"x{i+1}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i+1}", [-np.inf, np.inf]) for i in range(2)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names[1:]].to_numpy()
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x, axis=1)
        y1 = X["x1"].to_numpy()
        y2 = g * (1 - (y1 / g) ** 0.5 - (y1 / g) * np.sin(10 * np.pi * y1))
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)

    def get_optima(self, points=100):
        regions = [
            [0, 0.0830015349],
            [0.182228780, 0.2577623634],
            [0.4093136748, 0.4538821041],
            [0.6183967944, 0.6525117038],
            [0.8233317983, 0.8518328654],
        ]

        pf = []
        for r in regions:
            x1 = np.linspace(r[0], r[1], int(points / len(regions)))
            x2 = 1 - np.sqrt(x1) - x1 * np.sin(10 * np.pi * x1)
            pf.append(np.stack([x1, x2], axis=1))

        y = np.concatenate(pf, axis=0)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT4(Problem):
    """ZDT-4 benchmark problem."""

    def __init__(self, n_inputs=10):
        super().__init__(
            name="ZDT-4 problem",
            inputs=[Continuous("x1", [0, 1])]
            + [Continuous(f"x{i+1}", [-5, 5]) for i in range(1, n_inputs)],
            outputs=[Continuous(f"y{i+1}", [0, np.inf]) for i in range(2)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names].to_numpy()
        g = 1 + 10 * (self.n_inputs - 1)
        for i in range(1, self.n_inputs):
            g += x[:, i] ** 2 - 10 * np.cos(4.0 * np.pi * x[:, i])
        y1 = X["x1"].to_numpy()
        y2 = g * (1 - np.sqrt(y1 / g))
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)

    def get_optima(self, points=100):
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.sqrt(x)], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)


class ZDT6(Problem):
    """ZDT-6 benchmark problem."""

    def __init__(self, n_inputs=30):
        super().__init__(
            name="ZDT-6 problem",
            inputs=[Continuous(f"x{i+1}", [0, 1]) for i in range(n_inputs)],
            outputs=[Continuous(f"y{i+1}", [-np.inf, np.inf]) for i in range(2)],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = X[self.inputs.names].to_numpy()
        n = self.n_inputs
        g = 1 + 9 * (np.sum(x[:, 1:], axis=1) / (n - 1)) ** 0.25
        y1 = 1 - np.exp(-4 * x[:, 0]) * (np.sin(6 * np.pi * x[:, 0])) ** 6
        y2 = g * (1 - (y1 / g) ** 2)
        return pd.DataFrame({"y1": y1, "y2": y2}, index=X.index)

    def get_optima(self, points=100):
        x = np.linspace(0.2807753191, 1, points)
        y = np.stack([x, 1 - x**2], axis=1)
        return pd.DataFrame(y, columns=self.outputs.names)



import numpy as np
import pandas as pd

from opti.constraint import LinearInequality, NChooseK
from opti.objective import Maximize
from opti.parameter import Continuous, Discrete
from opti.problem import Problem


def _poly2(x: np.ndarray) -> np.ndarray:
    """Quadratic feature expansion including bias term."""
    return np.concatenate([[1], x, np.outer(x, x)[np.triu_indices(5)]])


class Detergent(Problem):
    """Detergent formulation problem.

    There are 5 outputs representing the washing performance on different stain types.
    Each output is modeled as a second degree polynomial.
    The formulation consists of 5 components.
    The sixth input is a filler (water) and is factored out and it's parameter bounds
    0.6 < water < 0.8 result in 2 linear inequality constraints for the other parameters.
    """

    def __init__(self):
        # coefficients for the 2-order polynomial; generated with
        # base = 3 * np.ones((1, 5))
        # scale = PolynomialFeatures(degree=2).fit_transform(base).T
        # coef = np.random.RandomState(42).normal(scale=scale, size=(len(scale), 5))
        # coef = np.clip(coef, 0, None)
        self.coef = np.array(
            [
                [0.4967, 0.0, 0.6477, 1.523, 0.0],
                [0.0, 4.7376, 2.3023, 0.0, 1.6277],
                [0.0, 0.0, 0.7259, 0.0, 0.0],
                [0.0, 0.0, 0.9427, 0.0, 0.0],
                [4.3969, 0.0, 0.2026, 0.0, 0.0],
                [0.3328, 0.0, 1.1271, 0.0, 0.0],
                [0.0, 16.6705, 0.0, 0.0, 7.4029],
                [0.0, 1.8798, 0.0, 0.0, 1.7718],
                [6.6462, 1.5423, 0.0, 0.0, 0.0],
                [0.0, 0.0, 9.5141, 3.0926, 0.0],
                [2.9168, 0.0, 0.0, 5.5051, 9.279],
                [8.3815, 0.0, 0.0, 2.9814, 8.7799],
                [0.0, 0.0, 0.0, 0.0, 7.3127],
                [12.2062, 0.0, 9.0318, 3.2547, 0.0],
                [3.2526, 13.8423, 0.0, 14.0818, 0.0],
                [7.3971, 0.7834, 0.0, 0.8258, 0.0],
                [0.0, 3.214, 13.301, 0.0, 0.0],
                [0.0, 8.2386, 2.9588, 0.0, 4.6194],
                [0.8737, 8.7178, 0.0, 0.0, 0.0],
                [0.0, 2.6651, 2.3495, 0.046, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        super().__init__(
            name="Detergent optimization",
            inputs=[
                Continuous("x1", domain=[0.0, 0.2]),
                Continuous("x2", domain=[0.0, 0.3]),
                Continuous("x3", domain=[0.02, 0.2]),
                Continuous("x4", domain=[0.0, 0.06]),
                Continuous("x5", domain=[0.0, 0.04]),
            ],
            outputs=[Continuous(f"y{i+1}", domain=[0, 3]) for i in range(5)],
            objectives=[Maximize(f"y{i+1}") for i in range(5)],
            constraints=[
                LinearInequality(["x1", "x2", "x3", "x4", "x5"], lhs=-1, rhs=-0.2),
                LinearInequality(["x1", "x2", "x3", "x4", "x5"], lhs=1, rhs=0.4),
            ],
        )

    def f(self, X: pd.DataFrame) -> pd.DataFrame:
        x = np.atleast_2d(X[self.inputs.names])
        xp = np.stack([_poly2(xi) for xi in x], axis=0)
        return pd.DataFrame(xp @ self.coef, columns=self.outputs.names, index=X.index)


class Detergent_NChooseKConstraint(Problem):
    """Variant of the Detergent problem where only 3 of the 5 formulation components are allowed to be active (n-choose-k constraint)."""

    def __init__(self):
        base = Detergent()

        super().__init__(
            name="Detergent optimization with n-choose-k constraint",
            inputs=base.inputs,
            outputs=base.outputs,
            objectives=base.objectives,
            constraints=list(base.constraints)
            + [NChooseK(names=base.inputs.names, max_active=3)],
            f=base.f,
        )


class Detergent_OutputConstraint(Problem):
    """Variant of the Detergent problem with an additional output/black-box constraint.

    In addition to the 5 washing performances there is a sixth output reflecting the stability of the formulation.
    If `discrete=True` the stability can only be measured qualitatively (0: not stable, 1: stable).
    If `discrete=True` the stability can be measured quantively with smaller values indicating less stable formulations.
    """

    def __init__(self, discrete=False):
        base = Detergent()

        def f(X):
            Y = base.f(X)
            if discrete:
                Y["stable"] = (X.sum(axis=1) < 0.3).astype(int)
            else:
                Y["stable"] = (0.4 - X.sum(axis=1)) / 0.2
            return Y

        outputs = list(base.outputs)
        if discrete:
            outputs += [Discrete("stable", domain=[0, 1])]
        else:
            outputs += [Continuous("stable", domain=[0, 1])]

        super().__init__(
            name="Detergent optimization with stability constraint",
            inputs=base.inputs,
            outputs=outputs,
            objectives=base.objectives,
            output_constraints=[Maximize("stable", target=0.5)],
            constraints=base.constraints,
            f=f,
        )


class Detergent_TwoOutputConstraints(Problem):
    """Variant of the Detergent problem with two outputs constraint.

    In addition to the 5 washing performances there are two more outputs measuring the formulation stability.
    The first, stability 1, measures the immediate stability. If not stable, the other properties cannot be measured, except for stability 2.
    The second, stability 2, measures the long-term stability.
    """

    def __init__(self):
        base = Detergent()

        def f(X: pd.DataFrame) -> pd.DataFrame:
            Y = base.f(X)
            x = self.get_X(X)
            stable1 = (x.sum(axis=1) < 0.3).astype(int)
            stable2 = (x[:, :-1].sum(axis=1) < 0.25).astype(int)
            Y[stable1 == 0] = np.nan
            Y["stability 1"] = stable1
            Y["stability 2"] = stable2
            return Y

        outputs = list(base.outputs) + [
            Discrete("stability 1", domain=[0, 1]),
            Discrete("stability 2", domain=[0, 1]),
        ]

        super().__init__(
            name="Detergent optimization with two output constraint",
            inputs=base.inputs,
            outputs=outputs,
            objectives=base.objectives,
            output_constraints=[
                Maximize("stability 1", target=0.5),
                Maximize("stability 2", target=0.5),
            ],
            constraints=base.constraints,
            f=f,
        )
